#
# comps.py: header list and component set (package groups) management
#
# Copyright 2002, 2003 Red Hat, Inc.
#
# This software may be freely redistributed under the terms of the GNU
# General Public License.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import rpm
import os, sys
from string import *
import urllib2
import time

from rhpl.log import log
from rhpl.translate import _, N_, _expandLang
import rhpl.comps
from size_string import size_string


NOT_INSTALLED = 0
INSTALLED = 1
TOINSTALL = 2
TOREMOVE = 3
TOUPGRADE = 4

PKGTYPE_MANDATORY = 0
PKGTYPE_DEFAULT = 1
PKGTYPE_OPTIONAL = 2

def listuniq(list):
    list.sort()
    ret = []
    for item in list:
        if item not in ret:
            ret.append(item)
    return ret


class Package:
    def __init__(self, hdr, method):
        self.h = hdr
        self.method = method
        self.installState = NOT_INSTALLED
        self.name = hdr[rpm.RPMTAG_NAME]
        self.filename = None

        self.oldState = None
    
    def __getitem__(self, item):
        if item == 1000005:
            return self.filename
        else:
            return self.h[item]

    def __repr__(self):
        return "%s" % (self.name,)

    def setFilename(self, filename):
        self.filename = filename

    def select(self, asDependency = 0):
        # this package may just be needed for a dependency and we want
        # to be able to revert to it if so (deps are selected prior to
        # a ts run and we then give people a chance to back out
        if asDependency == 1:
            if self.oldState is not None:
                print "WARNING: %s had an old state and tried to set it again.  %s to %s" % (self.name, self.oldState, self.installState)                
            else:
                self.oldState = self.installState

        if self.installState == INSTALLED:
            print "selecting already installed package %s" %(self.name,)
        elif self.installState == TOINSTALL or self.installState == TOUPGRADE:
            print "selecting already selected package %s" %(self.name,)
        elif self.installState == NOT_INSTALLED:
            self.installState = TOINSTALL
        elif self.installState == TOREMOVE:
            # FIXME: is this sane?  it can only be marked for removal
            # if it was previously installed afaics
            # We also lose the maybe upgrade state
            self.installState = INSTALLED

    def unselect(self, set = None, asDependency = 0):
        # this package may just be needed for a dependency and we want
        # to be able to revert to it if so (deps are selected prior to
        # a ts run and we then give people a chance to back out

        if asDependency == 1:
            if self.oldState is not None:
                print "WARNING: %s had an old state and tried to set it again.  %s to %s" % (self.name, self.oldState, self.installState)
            else:
                self.oldState = self.installState

        # if it was previously installed (which touprade also implies),
        # mark it for removal -- else, revert back to not installed
        if (self.installState == INSTALLED or self.installState == TOUPGRADE):
            self.installState = TOREMOVE
        elif self.installState == TOINSTALL:
            self.installState = NOT_INSTALLED
        elif self.installState == NOT_INSTALLED:
            print "Unselecting already uninstalled package %s" %(self.name,)
        elif self.installState == TOREMOVE:
            print "Unselecting %s already marked for removal" %(self.name,)

    def revert(self):
        if self.oldState is not None:
            self.installState = self.oldState
            self.oldState = None

    def isSelected(self, toInstallOnly = 0):
        if ((self.installState == TOINSTALL) or
            (self.installState == TOUPGRADE) or
            ((toInstallOnly == 0) and (self.installState == INSTALLED))):
            return 1
        else:
            return 0

    def isBeingRemoved(self):
        if self.installState == TOREMOVE:
            return 1
        else:
            return 0

    def isInstalled(self):
        return (self.installState == INSTALLED)

    def selectForUpgrade(self):
        self.installState = TOUPGRADE

    # this is a misnomer since we actually return the summary instead of
    # the description, but oh well
    def getDescription(self):
        return self.h[rpm.RPMTAG_SUMMARY]

    # FIXME: this is a quick hack to move the method into the package
    # so that we can install packages from multiple sources easily
    def getFilename(self, pkgTimer):
        return self.method.getFilename(self, pkgTimer)


class Component:
    def __init__(self, set, compgroup, doDeps = 1, langs = []):
        """set is the ComponentSet, compgroup is the rhpl.comps.Group"""
        
        self.set = set
        self.hidden = not compgroup.user_visible
        self.default = compgroup.default
        self.group = compgroup
        self.id = compgroup.id
        self.name = compgroup.name

        self.description = None
        self.displayName = None
        for lang in langs:
            if self.displayName is None and compgroup.translated_name.has_key(lang):
                self.displayName = compgroup.translated_name[lang]
            if (self.description is None and
                compgroup.translated_description.has_key(lang)):
                self.description = compgroup.translated_description[lang]
        # if we didn't find a translation, fall back to english
        if self.displayName is None:
            self.displayName = compgroup.name
        if self.description is None:
            self.description = compgroup.description

        self.pkgDict = {}
        self.installState = INSTALLED

        self.includes = []
        self.metapkgs = {}

        self.manuallySelected = 0

        packages = set.packages
        found = 0
        keys = compgroup.packages.keys()

        for pkg in keys:
            if not packages.has_key(pkg):
                log("%s references package %s which doesn't exist"
                    %(self.name, pkg))
                continue
            (type, name) = compgroup.packages[pkg]
            if type == u'mandatory':
                found = 1
                pkgtype = PKGTYPE_MANDATORY
                if not packages[pkg].isInstalled():
#                    print "%s threw out %s of being installed" %(pkg, self.name)
                    self.installState = NOT_INSTALLED
            elif type == u'default':
                pkgtype = PKGTYPE_DEFAULT
                if packages[pkg].isInstalled():
                    found = 1
            elif type == u'optional':
                pkgtype = PKGTYPE_OPTIONAL
                if packages[pkg].isInstalled():
                    found = 1
            else:
                log("Invalid package type of %s for %s in %s; default to optional" % (type, pkg, self.name))
                pkgtype = PKGTYPE_OPTIONAL

            self.addPackages(packages[pkg], pkgtype, doDeps)

        if found == 0 and self.installState == INSTALLED:
                   self.installState = NOT_INSTALLED


    def addPackages(self, pkg, pkgtype, handleDeps = 1):
        self.pkgDict[pkg] = (pkgtype,)
        
    def isInstalled(self):
        return (self.installState == INSTALLED)

    def addInclude(self, comp):
        self.includes.append(comp)

    def addMetaPkg(self, comp, isDefault = 0):
        if isDefault:
            self.metapkgs[comp] = (PKGTYPE_DEFAULT,)
        else:
            self.metapkgs[comp] = (PKGTYPE_OPTIONAL,)

    def getDescription(self):
        return self.description

    # return the total number of packages in the component
    def getTotalPackages(self):
        return len(self.pkgDict.keys()) + len(self.metapkgs.keys())

    # return the number of installed packages from the component
    def getNumberInstalledPackages(self):
        num = 0
        for pkg in self.pkgDict.keys():
            if pkg.isSelected() != 0:
                num = num + 1
        for metapkg in self.metapkgs.keys():
            if metapkg.isSelected():
                num = num + 1
        return num

    def filter_packages(self, func):
        rc = []
        for pkg in self.pkgDict.keys():
            if func(self.pkgDict[pkg]):
                rc.append(pkg)
        for pkg in self.metapkgs.keys():
            if func(self.metapkgs[pkg]):
                rc.append(pkg)
        return rc

    # sort packages by name
    def sortPackages(self, a, b):
        if a.name < b.name:
            return -1
        elif a.name > b.name:
            return 1
        else:
            return 0

    # return both the packages and the metapkgs
    def getPackages(self, type = None):
        rc = []
        if type is None:
            rc = self.pkgDict.keys()
            rc.extend(self.metapkgs.keys())
        elif type == PKGTYPE_MANDATORY:
            func = lambda pkg: (pkg[0] == PKGTYPE_MANDATORY)
            rc = self.filter_packages(func)
        # we want both the optional and defaults here, which is a little
        # silly, but c'est la vie
        elif type == PKGTYPE_OPTIONAL:
            func = lambda pkg: (pkg[0] == PKGTYPE_DEFAULT or
                                pkg[0] == PKGTYPE_OPTIONAL)
            rc = self.filter_packages(func)
        else:
            log("requested unknown package type %s" % (type,))
        rc.sort(self.sortPackages)
        return rc

    def select(self, forInclude = 0):
        if not forInclude:
            self.manuallySelected = 1
        if self.installState == INSTALLED or self.installState == TOINSTALL:
            return

        for comp in self.includes:
            comp.select(forInclude = 1)
        
        for comp in self.metapkgs.keys():
            (ptype,) = self.metapkgs[comp]
            if ptype == PKGTYPE_DEFAULT:
                comp.select(forInclude = 1)

        for pkg in self.pkgDict.keys():
            (ptype,) = self.pkgDict[pkg]
            # should we handle this differently?
            if ptype == PKGTYPE_DEFAULT or ptype == PKGTYPE_MANDATORY:
                self.selectOptionalPackage(pkg)
        
        if self.installState == INSTALLED:
            print "selecting already installed component %s" %(self.name,)
        elif self.installState == TOINSTALL:
            print "selecting already toinstall component %s" %(self.name,)
        elif self.installState == NOT_INSTALLED:
            self.installState = TOINSTALL
        elif self.installState == TOREMOVE:
            self.installState = INSTALLED

    def unselect(self, forInclude = 0):
        if not forInclude:
            self.manuallySelected = 0
        else:
            # FIXME: this isn't really right for an anaconda-type install
            # case, but we don't want people being able to traverse includes
            # by removing components from redhat-config-packages afaict
            # (it can lead to things like base being removed which is bad ;)
            return
        if self.installState == TOREMOVE or self.installState == NOT_INSTALLED:
            return

        for comp in self.includes:
            comp.unselect(forInclude = 1)

        for comp in self.metapkgs.keys():
            (ptype,) = self.metapkgs[comp]
            if ptype == PKGTYPE_DEFAULT:
                comp.unselect(forInclude = 1)

        for pkg in self.pkgDict.keys():
            (ptype,) = self.pkgDict[pkg]
            # should we handle this differently?
            pkg.unselect()


        if self.installState == INSTALLED:
            self.installState = TOREMOVE
        elif self.installState == TOINSTALL:
            self.installState = NOT_INSTALLED
        elif self.installState == NOT_INSTALLED:
            print "Unselecting already uninstalled package %s" %(self.name,)
        elif self.installState == TOREMOVE:
            print "Unselecting %s already marked for removal" %(self.name,)

    def selectOptionalPackage(self, pkg):
        pkg.select()

    def unselectOptionalPackage(self, pkg):
        pkg.unselect()
    
    def isSelected(self, justManual=0):
        if self.installState == INSTALLED or self.installState == TOINSTALL:
            return 1
        else:
            return 0


    # this is a bad hack, but components which are metapkgs sort of need this
    def __getitem__(self, item):
        if item == rpm.RPMTAG_NAME:
            return self.name
        elif item == rpm.RPMTAG_SIZE:
            size = 0
            for pkg in self.pkgDict.keys():
                size = size + pkg[rpm.RPMTAG_SIZE]
            return size
        elif item == rpm.RPMTAG_SUMMARY:
            return self.getDescription()

class ComponentSet(dict):
    def __init__(self, file, hdlist, arch = None, update_method = None):
        dict.__init__(self)
        self.frozen = 0
        self.archlist = []
        self.update_method = update_method
        if not arch:
            arch = os.uname()[4]

        self.archlist.append(arch)
        self.packages = hdlist
        
        self.instProvs = {}
        self.providesDict = {}
        self.requiresDict = {}

        self.readCompsFile(file, self.packages)

    # quick hack addition of a method...
    def setupStateDicts(self, packages):
        if self.update_method:
            apply (self.update_method, (0.5,))
        # figure out the installed state of our packages
        # FIXME: this really shouldn't be done here
        ts = rpm.TransactionSet("/")
        # screwy backwards logic
        ts.setVSFlags(~(rpm.RPMVSF_NORSA|rpm.RPMVSF_NODSA))
        ts.setFlags(~rpm.RPMTRANS_FLAG_NOMD5)

        keys = packages.keys()
        if self.update_method:
            incr = 0.5/len(keys)
            counter = 0.5
        for pkg in keys:
            if self.update_method:
                apply (self.update_method, (counter,))
                counter = counter + incr

            mi = ts.dbMatch('name', pkg)
            for h in mi:
                if packages[pkg].installState == NOT_INSTALLED:
                    packages[pkg].installState = INSTALLED

                # FIXME: these packages provide things that are needed
                # in some of the other packages.  they're kind of weird
                # like that so take them out of our fast path to try
                # to avoid problems
                if (pkg == "glibc-debug" or pkg == "kde2-compat"
                    or pkg.startswith("kernel-") or pkg == "qt2-Xt"):
                    continue

                # we can have packages without provides and it's then
                # None instead of [].  *sigh*
                if h[rpm.RPMTAG_PROVIDES] is None:
                    continue
                
                for prov in h[rpm.RPMTAG_PROVIDES]:
                    if self.instProvs.has_key(prov):
                        self.instProvs[prov].append(pkg)
                    else:
                        self.instProvs[prov] = [ pkg ]

        # now let's generate the requires dict so that we can try to
        # be intelligent on removals... again, this doesn't really
        # belong here
        mi = ts.dbMatch('name', pkg)
        for h in mi:
            name = h[rpm.RPMTAG_NAME]
            reqs = h[rpm.RPMTAG_REQUIRENAME]
            if reqs is None:
                reqs = []
            for req in reqs:
                if self.requiresDict.has_key(req):
                    self.requiresDict[req].append(name)
                else:
                    self.requiresDict[req] = [ name ]
        ts.closeDB()


    def readCompsFile(self, filename, packages):
        if filename is not None:
            connected = 0
            while not connected:
                try:
                    file = urllib2.urlopen(filename)
                except IOError, (errnum, msg):
                    log("IOError %s occurred getting %s: %s", filename,
                            errnum, str(msg))
                    time.sleep(5)
                else:
                    connected = 1
            self.compsxml = rhpl.comps.Comps(file)
            file.close()
        else:
            self.compsxml = rhpl.comps.Comps()

        self.setupStateDicts(packages)


        groups = self.compsxml.groups.keys()
        groups.sort()

        if os.environ.has_key("LANG"):
            langs = _expandLang(os.environ["LANG"])
        else:
            langs = []

        # first go through and make Comp object for
        # all of the groups
        for group in groups:
            group = self.compsxml.groups[group]
            comp = Component(self, group, langs = langs)
            self[comp.id] = comp
        for group in groups:
            group = self.compsxml.groups[group]
            comp = self[group.id]

            for id in group.groups.keys():
                if not self.has_key(id):
                    log("%s references component %s which doesn't exist"
                        %(group.name, id))
                    continue
                comp.addInclude(self[id])

            for id in group.metapkgs.keys():
                if not self.has_key(id):
                    log("%s references metapkg %s which doesn't exist"
                        %(group.name, id))
                    continue
                if group.metapkgs[id][0] == u'default':
                    comp.addMetaPkg(self[id], isDefault = 1)
                else:
                    comp.addMetaPkg(self[id], isDefault = 0)

    def sizeStr(self):
        return size_string (self.size())

    def size(self):
        size = 0
        for pkg in self.packages.values():
            if pkg.isSelected():
                size += (pkg[rpm.RPMTAG_SIZE])
        return size


    def revertState(self):
        """Reverts the state back to what it was before a deps check."""
        for pkg in self.packages.values():
            pkg.revert()

    def selectDeps (self, instPath, update_method = None):
        # faster dependency checking...
        # adding all the packages to a transaction set and
        # running depscheck is *slow*.  let's try to make it so
        # we only have to do that once

        update_method_counter = 0.0
        update_method_target = 0.0
        if update_method is not None:
            update_method (update_method_counter)

        tocheck = []
        for pkg in self.packages.values():
            if pkg.installState == TOINSTALL or pkg.installState == TOUPGRADE:
                tocheck.append(pkg)

        # we'll do a logarithmic approach to 0.67 here
        while len(tocheck) > 0:
            pkgs = tocheck
            tocheck = []

            update_method_counter = update_method_target
            update_method_target = (update_method_target +
                                    (0.67 - update_method_target)/2.0)
            counter = update_method_counter
            step = (update_method_target - update_method_counter) / len(pkgs)
            for pkg in pkgs:
                if update_method != None:
                    update_method (counter)
                    counter = counter + step

                reqs = pkg[rpm.RPMTAG_REQUIRENAME]
                if reqs is None:
                    reqs = []
                for req in reqs:
                    # ignore rpmlib deps
                    if req.startswith("rpmlib("):
                        continue
                    # if we have the req already, then don't worry
                    elif req in self.instProvs.keys():
                        continue
                    # if it's in the header list and not
                    # installed / to install, we need to make it to
                    # install and check it
                    elif req in self.packages.keys():
                        new = self.packages[req]
                    # look in our provides dict to see if we know where
                    # it is
                    elif req in self.providesDict.keys():
                        # the last thing in the list was the last one
                        # added, so we'll go with it since we don't really
                        # have any better heuristic :(
                        new = self.providesDict[req][-1]
                    else:
                        continue
                    if (new.installState == INSTALLED or
                        new.installState == TOINSTALL or
                        new.installState == TOUPGRADE):
                        continue
                    new.select(asDependency = 1)
                    tocheck.append(new)

        return update_method_target

    def unselectRemoveDeps (self, instPath, update_method = None):
        # removals need to be checked too so that we can set anything depending
        # on this to be removed.

        tocheck = []
        for pkg in self.packages.values():
            if pkg.installState == TOREMOVE:
                tocheck.append(pkg)

        while len(tocheck) > 0:
            pkgs = tocheck
            tocheck = []

            tmplist = []
            for pkg in pkgs:
                if self.requiresDict.has_key(pkg.name):
                    tmplist.extend(self.requiresDict[pkg.name])
                if pkg[rpm.RPMTAG_PROVIDES] is not None:
                    for prov in pkg[rpm.RPMTAG_PROVIDES]:
                        # if anything requires something which we provide
                        # and we're the only package that provides it,
                        # then we want to mark the package for removal
                        if (self.requiresDict.has_key(prov) and
                            self.instProvs.has_key(prov) and
                            len(self.instProvs[prov]) == 0):
                            tmplist.extend(self.requiresDict[prov])
                for fn in pkg[rpm.RPMTAG_FILENAMES]:
                    if self.requiresDict.has_key(fn):
                        tmplist.extend(self.requiresDict[fn])

            tmplist = listuniq(tmplist)
            for pkg in tmplist:
                pkg = self.packages[pkg]
                # if it's installed already, unconditionally select to
                # remove it
                if (pkg.installState == INSTALLED or
                    pkg.installState == TOUPGRADE):
                    tocheck.append(pkg)
                # if it's selected for being installed... unselect it for
                # now I guess.  we should probably pop up a warning dialog
                # in this case
                elif pkg.installState == TOINSTALL:
                    print "WARNING: %s was about to be installed, but got unselected as a dependency of something being removed." %(pkg.name,)
                    tocheck.append(pkg)
                else:
                    # we don't need to worry about these (not installed)
                    continue
                pkg.unselect(asDependency = 1)
            

    def verifyDeps (self, instPath, upgrade, update_method = None):
        update_method_counter = 0.0
        update_method_target = 0.0
        update_method (update_method_counter)

        def formatRequire (name, version, flags):
            string = name
            
            if flags:
                if flags & (rpm.RPMSENSE_LESS | rpm.RPMSENSE_GREATER | 
                            rpm.RPMSENSE_EQUAL):
                    string = string + " "
                    if flags & rpm.RPMSENSE_LESS:
                        string = string + "<"
                    if flags & rpm.RPMSENSE_GREATER:
                        string = string + ">"
                    if flags & rpm.RPMSENSE_EQUAL:
                        string = string + "="
                    string = string + " %s" % version
            return string

        if upgrade:
            how = 'u'
        else:
            how = 'i'

	checkDeps = 1
	rc = []
        extras = {}

        update_method_target = self.selectDeps(instPath, update_method)
        self.unselectRemoveDeps(instPath, update_method)
                    
        # update_method is done logarithmically.  That is, our first
        # pass through the while loop, we go from 0.0->0.5.  The next
        # pass, we go from 0.5->0.75.  It's very rough, and looks a
        # little odd when we only have one pass, but it should give a
        # reasonably moving progress bar for people to look at.
        #
        # Additionally, we give the first for loop 75% of the time,
        # and the deps check 25%.

	while checkDeps:
            update_method_counter = update_method_target
            update_method_target = update_method_target + (1.0 - update_method_target)/2.0
            counter = update_method_counter
            if upgrade:
                ts = rpm.TransactionSet(instPath, rpm.RPMVSF_NOHDRCHK)
                how = 'u'
            else:
                ts = rpm.TransactionSet()
                how = 'i'

            ts.setVSFlags(~(rpm.RPMVSF_NORSA|rpm.RPMVSF_NODSA))
            ts.setFlags(~rpm.RPMTRANS_FLAG_NOMD5)

            vals = self.packages.values()
            step = ((update_method_target - update_method_counter)*0.75) / len (vals)
            for p in vals:
                if update_method != None:
                    update_method (counter)
                    counter = counter + step
                if p.isSelected(toInstallOnly = 1):
                    ts.addInstall(p.h, (p.h, p.h[rpm.RPMTAG_NAME]), how)
                elif p.isBeingRemoved():
                    ts.addErase(p.h[rpm.RPMTAG_NAME])
                else:
                    if extras.has_key(p.h):
                        ts.addInstall(p.h, (p.h, p.h[rpm.RPMTAG_NAME]), how)
                    elif p.installState != INSTALLED:
                        ts.addInstall(p.h, (p.h, p.h[rpm.RPMTAG_NAME]), "a")
	    deps = ts.check()
	    checkDeps = 0
	    if not deps:
		break

            step = ((update_method_target - update_method_counter)*0.25) / len (deps)
            for ((name, version, release),
                 (reqname, reqversion),
                 flags, suggest, sense) in deps:
                if update_method != None:
                    update_method (counter)
                    counter = counter + step
                if sense == rpm.RPMDEP_SENSE_REQUIRES:
                    if suggest:
                        (header, sugname) = suggest
                        log ("depcheck: package %s needs %s (provided by %s)",
                             name, formatRequire(reqname, reqversion, flags),
                             sugname)
                        self.packages[sugname].select(asDependency = 1)
			checkDeps = 1
                    else:
                        log ("depcheck: package %s needs %s (not provided)",
                             name, formatRequire(reqname, reqversion, flags))
                        sug = (name, (reqname, reqversion, flags))
                        if not sug in rc:
                            rc.append (sug)
                elif sense == rpm.RPMDEP_SENSE_CONFLICTS:
		    # We need to check if the one we are going to
		    # install is ok.
		    conflicts = 1
		    if reqversion:
			fields = split(reqversion, '-')
			if (len (fields) == 2):
			    needed = ("", fields [0], fields [1])
			else:
			    needed = ("", fields [0], "")
                        try:
                            h = self.packages[reqname].h
                        except KeyError:
                            # we don't actually have the conflicting package
                            # in our available packages, the conflict is
                            # on the system.  Continue on.
                            continue
			installed = ("", h[rpm.RPMTAG_VERSION],
				     h [rpm.RPMTAG_RELEASE])
			if rpm.labelCompare (installed, needed) >= 0:
			    conflicts = 0

		    if conflicts:
			log ("%s-%s-%s conflicts with to-be-installed "
                             "package %s-%s, removing %s from set",
                             name, version, release, reqname, reqversion, reqname)
			if self.packages.has_key (reqname):
			    self.packages[reqname].unselect(asDependency = 1)
			    log ("... removed")
            
        del ts

        update_method (1.0)
        return rc
    


# this is a temporary way to set order of packages
def orderPackageGroups(curgroups):
    compsParents = curgroups.compsxml.hierarchy.order
    compsHierarchy = curgroups.compsxml.hierarchy

    grpids = []
    for grp in curgroups.values():
	grpids.append(grp.id)

    ignorelst = []
    retlist = []
    retdict = {}

    if os.environ.has_key("LANG"):
        langs = _expandLang(os.environ["LANG"])        
    else:
        langs = []

    for key in compsParents:
        # get the translated name 
        myname = None
        if not compsHierarchy.translations.has_key(key):
            myname = key
        else:
            for lang in langs:
                if compsHierarchy.translations[key].has_key(lang):
                    myname = compsHierarchy.translations[key][lang]
                    break
            if myname is None:
                myname = key
        
        retlist.append(myname)
        retdict[myname] = []
        
	compslist = compsHierarchy[key]
	for grp in compslist:

	    if grp in grpids:
		thecomp = curgroups[grp]
		ignorelst.append(grp)
                retdict[myname].append(thecomp)

    miscgrp = _("Miscellaneous")
    for grp in grpids:
	if grp in ignorelst:
	    continue

	thecomp = curgroups[grp]
        if thecomp.hidden != 0:
            continue
	if miscgrp not in retlist:
	    retlist.append(miscgrp)
	    retdict[miscgrp] = [thecomp]
	else:
	    retdict[miscgrp].append(thecomp)

    return (retlist, retdict)

def getCompGroupDescription(comp):
    if comp.name == u"Everything":
	return _("This group includes all the packages available.  Note that "
		 "this is substantially more packages than just the ones "
		 "in all the other package groups on this page.")
    
    descr = comp.description
    if descr:
	return _(descr)
    else:
	return None
